# THIS SOFTWARE IS PROVIDED "AS IS" AND WITHOUT ANY EXPRESS OR
# IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
# WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
# Name:           UVprojection.rb 
# Author:         Jan Sandstrom   www.pixero.com
# Date:           27 january 2011
# Description:	  Use for projecting planar UVs on selection. 
#                 Works with both ungrouped and grouped geometry.
# Based on Whaat's UVtools script
### TIG tweaks shown with '###'
# Thanks to Thomthom for UV samples code.

require 'sketchup.rb'

#Sketchup.send_action "showRubyPanel:"

if not file_loaded?(__FILE__)

	ents=Sketchup.active_model.selection ###

	menu = UI.menu("Plugins").add_submenu("UV Projection")
	menu.add_item("Planar"){UVprojection.new.planar_map(ents)}

	UI.add_context_menu_handler {|menu|
		sel=Sketchup.active_model.selection.first
		sub=menu.add_submenu("UV Projection")
		sub.add_item("Planar") {(UVprojection.new.planar_map(ents))}
	}
end
file_loaded(__FILE__)


class UVprojection

def initialize

	@points=[]
	@box=nil
		
	# Planar Dialog
	@axis = "Z  (blue)" if not @axis
	@tex_size = "Keep Current" if not @tex_size ###
	@tileW = 1.0 if not @tileW
	@tileH = 1.0 if not @tileH
	@influence = "Local" if not @influence
	@tex_width = 10.inch if not @tex_width ###
	@tex_height = 10.inch if not @tex_height ###
	@proj_side = "Front" if not @proj_side

	enums = ["X  (red)|Y  (green)|Z  (blue)", "Keep Current|Fit|Set Tiling|Set Size", "", "", "Local|Global", "", "", "Front|Back|Both"]
	prompts = ["Projection Axis: ", "Texture Size: ", "Tiling Width: ", "Tiling Height: ", "Influence: ", "Texture Width: ", "Texture Height: ", "Project UVs on: "]
	values = [@axis, @tex_size, @tileW, @tileH, @influence, @tex_width, @tex_height, @proj_side]
	results = inputbox(prompts, values, enums, "Planar Settings")
    return nil if not results
	@axis, @tex_size, @tileW, @tileH, @influence, @tex_width, @tex_height, @proj_side = results

end	
	
######################
def get_bounds(ents) # Get the bounds of the selection. Used in get_max_min to determine the max and min of the projection.

all_points=[]

definitions = []
# Iterate entities
		ents.each { |e|
			if e.kind_of?(Sketchup::ComponentInstance) || e.kind_of?(Sketchup::Group)
				parent = e.definition if e.kind_of?(Sketchup::ComponentInstance)
				parent = e.entities.parent if e.kind_of?(Sketchup::Group)				
				next if definitions.include?(parent)					
				parent.entities.each { |ents|
					if (ents.class==Sketchup::Face)
						# Apply the inverse of the Group/Components transformation to all grouped vertices
						t = parent.transformation.invert!  									
						ents.vertices.transform t
						all_points.push(ents.vertices)
					end
				}
				definitions.push(parent)
			elsif (e.class==Sketchup::Face)
				all_points.push(e.vertices)				
			end
		}

return if all_points.empty?

all_points.flatten!
all_points.uniq!

box=Geom::BoundingBox.new
all_points.each {|v| box.add(v.position)}

@box=box

end
############################
def get_faces(ents) #Get the faces to get the vertices.
faces = []
definitions = []
# Iterate selected entities
		ents.each { |e|
			if e.kind_of?(Sketchup::ComponentInstance) || e.kind_of?(Sketchup::Group)
				parent = e.definition if e.kind_of?(Sketchup::ComponentInstance)
				parent = e.entities.parent if e.kind_of?(Sketchup::Group)				
				next if definitions.include?(parent)					
				parent.entities.each { |ents|
					if ents.class==Sketchup::Face 
						faces.push(ents)
					end
				}
				definitions.push(parent)
			elsif e.class==Sketchup::Face 
				faces.push(e)
			end
		}

return faces 

end
######################

def get_max_min(ents)  #Get the max and min for the boundingbox of the selection.
	get_bounds(ents)	
		return [@box.max.x, @box.min.x, @box.max.y, @box.min.y, @box.max.z, @box.min.z]	

end
########################
def resize_uvs(uvs, mat) #For changing/setting the size of the texture

tex=mat.texture
return if tex==nil

new_uvs=[]

if (@tex_size == "Keep Current")
	if (@axis == "X  (red)")
		y_tile=(@box.max.y-@box.min.y)/tex.width
		z_tile=(@box.max.z-@box.min.z)/tex.height
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*y_tile
			new_uvs[i].y=uvs[i].y*z_tile
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Y  (green)")
		x_tile=(@box.max.x-@box.min.x)/tex.width
		z_tile=(@box.max.z-@box.min.z)/tex.height
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*x_tile
			new_uvs[i].y=uvs[i].y*z_tile
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Z  (blue)")
		x_tile=(@box.max.x-@box.min.x)/tex.width
		y_tile=(@box.max.y-@box.min.y)/tex.height
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*x_tile
			new_uvs[i].y=uvs[i].y*y_tile
			new_uvs[i].z=uvs[i].z
		}
	end
end
if (@tex_size == "Fit")	
	if (@axis == "X  (red)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x
			new_uvs[i].y=uvs[i].y
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Y  (green)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x
			new_uvs[i].y=uvs[i].y
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Z  (blue)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x
			new_uvs[i].y=uvs[i].y
			new_uvs[i].z=uvs[i].z
		}
	end
end
if (@tex_size == "Set Tiling")
	if (@axis == "X  (red)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*@tileW
			new_uvs[i].y=uvs[i].y*@tileH
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Y  (green)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*@tileW
			new_uvs[i].y=uvs[i].y*@tileH
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Z  (blue)")
		uvs.each_index {|i|
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*@tileW
			new_uvs[i].y=uvs[i].y*@tileH
			new_uvs[i].z=uvs[i].z
		}
	end
end

if (@tex_size == "Set Size")

	if (@axis == "X  (red)")
		uvs.each_index {|i|
			tw = (@box.max.y-@box.min.y)/(@tex_width)
			th = (@box.max.z-@box.min.z)/(@tex_height)
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*tw
			new_uvs[i].y=uvs[i].y*th
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Y  (green)")
		uvs.each_index {|i|
			tw = (@box.max.x-@box.min.x)/(@tex_width)
			th = (@box.max.z-@box.min.z)/(@tex_height)
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*tw
			new_uvs[i].y=uvs[i].y*th
			new_uvs[i].z=uvs[i].z
		}
	end
	if (@axis == "Z  (blue)")
		uvs.each_index {|i|
			tw = (@box.max.x-@box.min.x)/(@tex_width)
			th = (@box.max.y-@box.min.y)/(@tex_height)
			new_uvs[i]=Geom::Point3d.new(0,0,0)
			new_uvs[i].x=uvs[i].x*tw
			new_uvs[i].y=uvs[i].y*th
			new_uvs[i].z=uvs[i].z
		}
	end
	
		#Set texture size for the material which will affect all entities with that texture.		
	if (@influence == "Global") 
		tex.size = [@tex_width, @tex_height]###
	end

end

	return new_uvs

end

##############################
def get_uv_plane(extents)	
	x = Geom::Vector3d.new(1,0,0)
	y = Geom::Vector3d.new(0,1,0)
	z = Geom::Vector3d.new(0,0,1)
	
	@origo = Geom::Point3d.new(0, 0, 0)

	if (@axis == "X  (red)")	
		return [@origo, x]
	end
	if (@axis == "Y  (green)")	
		return [@origo, y]
	end
	if (@axis == "Z  (blue)")
		return [@origo, z]
	end
end

#########################
def get_planar_UV(point3d, face, extents)
    # Transformation to convert 3d points on the plane to local 2d points.
	@normal = @plane[1]
	@xaxis = @normal.axes.x
	@origin = @origo
    o = @origin
	x = @xaxis
    y = @normal * @xaxis
    @to_local = Geom::Transformation.axes(o, x, y, @normal).inverse
	
	xmax=extents[0]
	xmin=extents[1]       
	ymax=extents[2]
	ymin=extents[3]        
	zmax=extents[4]
	zmin=extents[5]
	
	xmaxmin = xmax-xmin
	ymaxmin = ymax-ymin
	zmaxmin = zmax-zmin
		
	# Divide by zero catch
	if xmaxmin < 0.001
		xmaxmin = 0.001
	end
	if ymaxmin < 0.001
		ymaxmin = 0.001
	end
	if zmaxmin < 0.001
		zmaxmin = 0.001
	end

    point2d = point3d.transform(@to_local)
	
	if (@axis == "X  (red)")	
		u = (point2d.x-ymin)/(ymaxmin)
		v = (point2d.y-zmin)/(zmaxmin)
	elsif (@axis == "Y  (green)")	
		u = (point2d.x-xmin)/(xmaxmin)
		v = (point2d.y-zmin)/(zmaxmin)
	elsif (@axis == "Z  (blue)")	
		u = (point2d.x-xmin)/(xmaxmin)
		v = (point2d.y-ymin)/(ymaxmin)
	end

	return [u, v, 1.0]
end

#########################
def to_a
    return [origin, normal]
end
#########################
def planar_map(ents)  #Performs planar mapping

	Sketchup.active_model.start_operation("Planar")
	@faces=get_faces(ents)
	extents=get_max_min(@faces)

	@faces.each {|face|
		material=face.material
		@active_mat = material
		@width = @active_mat.texture.width
		@height = @active_mat.texture.height
		
		#### UV sampling code borrowed from Thomthom:
		# The UV mapping co-ords to feed position_material.  
		# Can't rely on the vertex positions, since a face might have only three
		# and four sets are required to correctly map a distorted texture.
		# And the point can't be co-linear either.
		#
		# Instead, generate four points on the plane of the target face.
		# 3  4
		#
		# 1  2
		pts = []
		pts << face.vertices.first.position               	 # 1
		pts << pts.first.offset( face.normal.axes.x, 10.mm ) # 2
		pts << pts.first.offset( face.normal.axes.y, 10.mm ) # 3
		pts << pts.last.offset(  face.normal.axes.x, 10.mm ) # 4
		####
		
		uv_points = []
		uv = []
		@points = []
		@uvs = []
		# Get the UV projection plane.
		@plane = get_uv_plane(extents)

		# Check to see if a face normal is exactly perpendicular to the projection.
		# If it is, it will require a special "treatment".	
		if (face.normal.perpendicular? @plane[1])
			# Do special mapping in face space with u or v values stretched across face.
			face_plane=[face.vertices.first.position, face.normal]
			p_a = []
			uvs_a = []
			offset = 0.0001.mm # A very small offset value.
			
			# Project three points to the projection plane and get the UV for them.
			# Since a perpendicular face have no area in the projection direction we will use two points for U (points 1 and 3) or two for V (points 1 and 2) to get uvs. 
			# We will then copy them and add a small offset and use these new offsetUV's for the remaining two points UVs. 

			perp_proj_array=[pts[0],pts[1],pts[2]] 
			
			perp_proj_array.each { |p1|
				# Get UV coordinates.
				p2 = p1.project_to_plane(@plane.to_a)
				uv = get_planar_UV(p2, face, extents)
				# Add the points to the pool.
				p_a << p1
				uvs_a << uv
			}
			# Now get the points (in face_plane) that will be used for mapping.
			pts.each { |p1|
				# Get UV coordinates.
				p2 = p1.project_to_plane(face_plane)
				uv = get_planar_UV(p2, face, extents)
				# Add the points to the pool.
				@points << p1
				@uvs << uv
			}
			# Add the offset
			offsetUV2=[uvs_a[0].x+offset, uvs_a[0].y+offset, 1]
			offsetUV4X=[uvs_a[2].x+offset, uvs_a[2].y+offset, 1]
			offsetUV4Y=[uvs_a[1].x+offset, uvs_a[1].y+offset, 1]
			# Depending on which axis is parallel, use different UVs.
			if (face.normal.axes.x.parallel? @plane[1])
				@uvs[0]=uvs_a[0]
				@uvs[1]=offsetUV2 #uv2
				@uvs[2]=uvs_a[2]
				@uvs[3]=offsetUV4X #uv4				
			else
				@uvs[0]=uvs_a[0]
				@uvs[1]=uvs_a[1]
				@uvs[2]=offsetUV2 #uv3
				@uvs[3]=offsetUV4Y #uv4
			end
				
		else	# For all other faces that are NOT exactly perpendicular to the projection.					
			pts.each { |p1|
				# Get UV coordinates.
				p2 = p1.project_to_plane(@plane.to_a)
				uv = get_planar_UV(p2, face, extents)
				# Add the points to the pool.
				@points << p1
				@uvs << uv
			}					
		end

		# Resize UVs.
		@uvs = resize_uvs(@uvs, material)
		
		# Add to the uv_points array.
		uv_points=[@points[0],@uvs[0],@points[1],@uvs[1],@points[2],@uvs[2],@points[3],@uvs[3]]

		# Position texture.
		if (@proj_side == "Front")
			face.position_material(material, uv_points, true) # for frontface
		elsif (@proj_side == "Back")
			face.position_material(material, uv_points, false) # for backface
		elsif (@proj_side == "Both")
			face.position_material(material, uv_points, true) # for frontface
			face.position_material(material, uv_points, false) # for backface
		end
	}
Sketchup.active_model.commit_operation
end	

end #class

